/*
 * Source File < graphics/color.c >
 *
 * Standard Functions: Borland Graphics Interface for Xlib
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#include "_bgi.h"
char BGISCOLORNAME[MAXCOLORS][64] = {
#include "colors/scolor.def"
};

/*
 * standard functions to get and set color
 */
int getcolor( void )
{
  return XBGI->color;
}
int getbkcolor( void )
{
  return XBGI->bkcolor;
}
int getmaxcolor( void )
{
  return XBGI->palette.size-1;
}
void setcolor( int color )
{
  XBGI->color = minmax( 0, color, XBGI->palette.size-1 );
  XSetForeground( XBGI->dpy, XBGI->gcdraw, XBGI->pixels[XBGI->color] );
}
void setbkcolor( int color )
{
  XBGI->bkcolor = minmax( 0, color, XBGI->palette.size-1 );
  XSetWindowBackground( XBGI->dpy, XBGI->win[XBGI->apage], 
                       XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcdraw,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcfill,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcpixel, XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcimage, XBGI->pixels[XBGI->bkcolor] );
  /*XClearWindow( XBGI->dpy, XBGI->win[XBGI->apage] );*/
}
void setmaxcolor( int c )
{
  int i, cc;
  unsigned long plane_mask;
  struct rgbtype {
    unsigned short red, green, blue;
  } rgb[8] = {
    { 0, 0, 0 },
    { 0, 0, 1 },
    { 0, 1, 0 },
    { 0, 1, 1 },
    { 1, 0, 0 },
    { 1, 0, 1 },
    { 1, 1, 0 },
    { 1, 1, 1 },
  };
  XColor exact, color;
  Status ok;

  switch( XBGI->gmode ){
  case BGIMONO:
  case BGISGRAY: /* It's scamping. */
    XBGI->palette.size = 2;
    XBGI->colortable[BlackPixel(XBGI->dpy,XBGI->scr)] = BLACK;
    XBGI->colortable[WhitePixel(XBGI->dpy,XBGI->scr)] = WHITE;
    XBGI->pixels[BLACK] = BlackPixel( XBGI->dpy, XBGI->scr );
    XBGI->pixels[WHITE] = WhitePixel( XBGI->dpy, XBGI->scr );
    for ( i=0; i<XBGI->palette.size; i++ )
      XBGI->palette.colors[i] = i;
    break;
  case BGIGRAY: /* I don't know grayscale visual class. */
    XBGI->palette.size = c + 1;
    ok = XAllocColorCells( XBGI->dpy, XBGI->colormap,
                          False, &plane_mask, False,
                          XBGI->pixels, XBGI->palette.size );
    if ( ok == False ) FatalError( "gray-scale cell allocation error" );
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    for ( i=1; i<XBGI->palette.size; i++ )
      setrgbpalette( i, i, i, i );
    setallpalette( &XBGI->palette );
    break;
  case BGISCOLOR: /* It's supported since version 2.2. */
    if ( c<1 || 255<c ) return;
    XBGI->palette.size = c + 1;
    for ( i=0; i<XBGI->palette.size; i++ ){
      ok = XAllocNamedColor( XBGI->dpy, XBGI->colormap,
                            BGISCOLORNAME[i], &exact, &color );
      if ( ok == False ) Warning( "named color allocation fault" );
      else{
        XBGI->pixels[i] = exact.pixel;
        XBGI->colortable[XBGI->pixels[i]] = i;
        XBGI->palette.colors[i] = i;
      }
    }
    break;
  case BGICOLOR:
    if ( c<1 || 255<c ) return;
    if ( XBGI->gerror != grNoInitGraph )
      XFreeColors( XBGI->dpy, XBGI->colormap, XBGI->pixels,
                  XBGI->palette.size, 0 );
    XBGI->palette.size = c + 1;
    ok = XAllocColorCells( XBGI->dpy, XBGI->colormap,
                          False, &plane_mask, False,
                          XBGI->pixels, XBGI->palette.size );
    if ( ok == False ){
      /* 1st failed */
      XBGI->colormap = XCreateColormap( XBGI->dpy, XBGI->root, XBGI->vis,
                                       AllocNone );
      /*XInstallColormap( XBGI->dpy, XBGI->colormap );*/ 
      XSetWindowColormap( XBGI->dpy, XBGI->root, XBGI->colormap );
      ok = XAllocColorCells( XBGI->dpy, XBGI->colormap,
                            False, &plane_mask, False,
                            XBGI->pixels, XBGI->palette.size );
      /* need technique for settings near the standard color map */
      if ( ok == False ) FatalError( "color cell allocation fault" );
    }
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    if ( XBGI->colornames && XBGI->gerror == grNoInitGraph ){
      /* color name setting at graph booting */
      for ( i=0; i<XBGI->palette.size; i++ ){
        ok = XParseColor( XBGI->dpy, XBGI->colormap,
                         BGISCOLORNAME[i], &color );
        if ( ok == False ) Warning( "named color storing fault" );
        else
          _setrgbpalette( i, color.red, color.green, color.blue );
      }
    }else{
      /* normal algorithum color setting */
      for ( i=0; i<XBGI->palette.size/8; i++ )
        for ( cc=0; cc<8; cc++ )
          setrgbpalette( 8*i+cc,
                        8*(i+1)*rgb[cc].red,
                        8*(i+1)*rgb[cc].green,
                        8*(i+1)*rgb[cc].blue );
      for ( i=1; i<XBGI->palette.size/8; i++ )
        setrgbpalette( 8*i+0, 4*(i+1), 4*(i+1), 4*(i+1) );
      cc = XBGI->palette.size/2;
      for ( i=0; i<XBGI->palette.size/8; i++ )
        setrgbpalette( 8*i+7, cc+4*(i+1), cc+4*(i+1), cc+4*(i+1) );
    }
    break;
  }
}
