/*
 * Source File < graphics/driver.c >
 *
 * Standard Functions: Borland Graphics Interface for Xlib
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#include "_bgi.h"
#include "option.h"

/*
 * registered enhanced function for graphics driver and mode
 */
char* getdriversname( int driver_number )
{
  static char *drivers_name[BGI+1] = {
    "DETECT",
    "BGI",
  };

  switch( driver_number ){
  case DETECT:
  case BGI:
    return drivers_name[driver_number];
  }
  return NULL;
}
/*
 * standard functions for graphics driver and mode
 */
char* getdrivername( void )
{
  return getdriversname( XBGI->gdriver );
}
void setgraphmode( int gmode )
{
  XBGI->gmode = gmode;
  switch( XBGI->gmode ){
  case BGIMONO:
  case BGISGRAY:
    setmaxcolor( 2 );
    break;
  case BGIGRAY:
  case BGISCOLOR:
  case BGICOLOR:
    setmaxcolor( XBGI->palette.size );
    break;
  }
}
int getgraphmode( void )
{
  return XBGI->gmode;
}
char* getmodename( int mode_number )
{
  static char *mode_name[BGICOLOR+1] = {
    "BGIMONO",
    "BGISGRAY",
    "BGIGRAY",
    "BGISCOLOR",
    "BGICOLOR",
  };

  switch( mode_number ){
  case BGIMONO:
  case BGISGRAY:
  case BGIGRAY:
  case BGISCOLOR:
  case BGICOLOR:
    return mode_name[mode_number];
  }
  return NULL;
}
int getmaxmode( void )
{
  return BGICOLOR;
}
void getmoderange( int graphdriver, int *lomode, int *himode )
{
  NotSupported( "getmoderange()" );
}
void _graphfreemem( void *ptr, unsigned size )
{
  NotSupported( "_graphfreemem()" );
}
void* _graphgetmem( unsigned size )
{
  NotSupported( "_graphgetmem()" );
  return NULL;
}
unsigned setgraphbufsize( unsigned bufsize )
{
  NotSupported( "setgraphbufsize()" );
  return 0;
}
int installuserdriver( char *name, int (*detect)( void ) )
{
  ReservedFuture( "installuserdriver()" );
  return 0;
}
int registerbgidriver( void (*driver)( void ) )
{
  XBGI_CURRENT_DRIVER = driver;
  return grOk;
}
int registerfarbgidriver( void *driver )
{
  XBGI_CURRENT_DRIVER = driver;
  return grOk;
}
void detectgraph( int *graphdriver, int *graphmode )
{
  /*
   * case of detecting before initgraph, always BGI and BGICOLOR
   */
  if ( XBGIS == NULL ){
    *graphdriver = BGI;
    *graphmode   = BGICOLOR;
    return;
  }
  /*
   * case of detecting after initgraph, research visual class and depth
   */
  XBGI->depth  = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->vis    = DefaultVisual( XBGI->dpy, XBGI->scr );
  XBGI->vclass = XBGI->vis->class;
  switch( *graphdriver ){
  case DETECT: default:
    *graphdriver = XBGI->gdriver = BGI;
    switch ( XBGI->vclass ){
    case StaticGray:
      *graphmode = XBGI->gmode = BGISGRAY;
      if ( XBGI->depth == 1 ){
        *graphmode = XBGI->gmode = BGIMONO;
	XBGI->ncolor = 2;
	XBGI->palette.size = (unsigned char)(XBGI->ncolor-1);	
      }
      break;
    case GrayScale:
      *graphmode = XBGI->gmode = BGIGRAY;
      if ( XBGI->depth == 1 ){
        *graphmode = XBGI->gmode = BGIMONO;
	XBGI->ncolor = 2;
	XBGI->palette.size = (unsigned char)(XBGI->ncolor-1);	
      }
      break;
    case StaticColor:
      *graphmode = XBGI->gmode = BGISCOLOR;
      break;
    case PseudoColor:
    case TrueColor:
    case DirectColor:
      *graphmode = XBGI->gmode = BGICOLOR;
      break;
    }
    break;
  case BGI:
    *graphmode = XBGI->gmode;
    /* no check */
    break;
  }
}
/*
 * default special driver, destroyer and creater
 */
void _BGI_driver( void )
{
  XUnmapSubwindows( XBGI->dpy, XBGI->root );
  XUnmapWindow( XBGI->dpy, XBGI->root );
  XFlush( XBGI->dpy );

  XFreeGC( XBGI->dpy, XBGI->gcdraw  );
  XFreeGC( XBGI->dpy, XBGI->gcfill  );
  XFreeGC( XBGI->dpy, XBGI->gcpixel );
  XFreeGC( XBGI->dpy, XBGI->gcimage );

  XFreeColormap( XBGI->dpy, XBGI->colormap );
  if ( XBGI->fs ) XFreeFont( XBGI->dpy, XBGI->fs );

  XDestroySubwindows( XBGI->dpy, XBGI->root );
  XDestroyWindow( XBGI->dpy, XBGI->root );

#if 0 /* for that events are remaining */
  XSync( XBGI->dpy, False );
  XCloseDisplay( XBGI->dpy );
#endif
}
void BGI_driver( void )
{
#include "bitmaps/icons/xbgi_a.xbm"
#include "bitmaps/icons/xbgi_b.xbm"
#include "bitmaps/icons/xbgi_c.xbm"
  struct xbm {
    char *bits;
    int width, height;
  } xbms[] = {
    { xbgi_c_bits, xbgi_c_width, xbgi_c_height },
    { xbgi_b_bits, xbgi_b_width, xbgi_b_height },
    { xbgi_a_bits, xbgi_a_width, xbgi_a_height },
    { NULL, 0, 0 }
  };
  int i, j, c, flag_mask, status;
  XSizeHints def;
  XSetWindowAttributes att;
  XEvent ev;
  XColor fgexact, fgcolor, bgexact, bgcolor, lcolor, dcolor;
  Pixmap bm;
  Status ok;
  extern char BGISCOLORNAME[MAXCOLORS][64];
  int argc = XBGI->argc;
  char **argv = XBGI->argv;

  char displayname[64] = "";
  char fnname[80]      = "lucidasanstypewriter-bold-12";
  char fgname[20]      = "white";
  char bgname[20]      = "gray50";
  char defgeomstr[20]  = "640x400";
  char usrgeomstr[20]  = "640x400";
  char titlename[80]   = "";
  char iconname[80]    = "";
  char iconfile[80]    = "";
  
  XBGI->ncolor  = 16;
  XBGI->color   = XBGI->ncolor-1;
  XBGI->bkcolor = 0;
  XBGI->npage   = 2;
  XBGI->apage   = 0;
  XBGI->vpage   = 0;
  strcpy( titlename, argv[0] );
  strcpy( iconname,  argv[0] );
  XBGI->only_ascii = 0;
  XBGI->colornames = 0;

  /* get display name and initialization */
  scan_option( argc, argv, "-display %s", displayname );
  XBGI->dpy = XOpenDisplay( displayname );
  if ( XBGI->dpy == NULL ) FatalError( "X cannot connect to display." );
  XBGI->scr = DefaultScreen( XBGI->dpy );
  XBGI->colormap = DefaultColormap( XBGI->dpy, XBGI->scr );

  /* position, size, gc, font and color settings for work */
  scan_option( argc, argv, "-geometry %s", usrgeomstr );
  scan_option( argc, argv, "-fn %s", fnname );
  scan_option( argc, argv, "-fg %s", fgname );
  scan_option( argc, argv, "-bg %s", bgname );
  XBGI->gc = XCreateGC( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr), 0, NULL );
  XBGI->fs = XLoadQueryFont( XBGI->dpy, fnname );
  if ( XBGI->fs == NULL ) Warning( "X cannot find font structure." );
  else
    XSetFont( XBGI->dpy, XBGI->gc, XBGI->fs->fid );
  ok = XAllocNamedColor( XBGI->dpy, XBGI->colormap, fgname,
                        &fgexact, &fgcolor );
  XBGI->fg = (ok==True)?fgexact.pixel:WhitePixel(XBGI->dpy,XBGI->scr);
  ok = XAllocNamedColor( XBGI->dpy, XBGI->colormap, bgname,
                        &bgexact, &bgcolor );
  XBGI->bg = (ok==True)?bgexact.pixel:BlackPixel(XBGI->dpy,XBGI->scr);
  lcolor.red   = min( bgexact.red  *5/4, 65535 );
  lcolor.green = min( bgexact.green*5/4, 65535 );
  lcolor.blue  = min( bgexact.blue *5/4, 65535 );
  dcolor.red   = max( 0, bgexact.red  *3/4 );
  dcolor.green = max( 0, bgexact.green*3/4 );
  dcolor.blue  = max( 0, bgexact.blue *3/4 );
  ok = XAllocColor( XBGI->dpy, XBGI->colormap, &lcolor );
  XBGI->lc = (ok==True)?lcolor.pixel:BlackPixel(XBGI->dpy,XBGI->scr);
  ok = XAllocColor( XBGI->dpy, XBGI->colormap, &dcolor );
  XBGI->dc = (ok==True)?dcolor.pixel:BlackPixel(XBGI->dpy,XBGI->scr);
  XSetForeground( XBGI->dpy, XBGI->gc, XBGI->fg );

  /* titlename and iconname settings */
  if ( scan_option( argc, argv, "-title %s", titlename ) )
    strcpy( iconname, titlename );
  scan_option( argc, argv, "-name %s", iconname );
  scan_option( argc, argv, "-icon %s", iconfile );

  /* get user bgi settings */
  scan_option( argc, argv, "-gd %d",      &(XBGI->gdriver) );
  scan_option( argc, argv, "-gm %d",      &(XBGI->gmode)   );
  scan_option( argc, argv, "-ncolor %d",  &(XBGI->ncolor)  );
  scan_option( argc, argv, "-color %d",   &(XBGI->color)   );
  scan_option( argc, argv, "-bkcolor %d", &(XBGI->bkcolor) );
  scan_option( argc, argv, "-npage %d",   &(XBGI->npage)   );
  scan_option( argc, argv, "-apage %d",   &(XBGI->apage)   );
  scan_option( argc, argv, "-vpage %d",   &(XBGI->vpage)   );

  /* scanning named color list */
  if ( (j=scan_option(argc,argv,"-colornames")) ){
    c = 0;
    for ( i=j+1; i<argc; i++ ){
      if ( *argv[i] == '-' )
        break;
      strcpy( BGISCOLORNAME[c++], argv[i] );
    }
    if ( 2 <= c && c<256 ){
      XBGI->colornames = !0;
      XBGI->ncolor = c;
    }
  }
  XBGI->palette.size = (unsigned char)(XBGI->ncolor-1);

  /* set font valuable */
  if ( scan_option( argc, argv, "-only_ascii"  ) ) XBGI->only_ascii = !0;
  if ( scan_option( argc, argv, "-only_ascii-" ) ) XBGI->only_ascii =  0;
#ifdef ENABLE_MULTILINGUAL
  if ( setlocale( LC_ALL, "" ) == NULL )
    XBGI->only_ascii = !0;
  else if ( XSupportsLocale() == False )
    XBGI->only_ascii = !0;
#endif

  /* create windows */
  flag_mask = XGeometry( XBGI->dpy, XBGI->scr, usrgeomstr, defgeomstr,
                        0, 1, 1, 0, 0,
                        &def.x, &def.y, &def.width, &def.height );
  XBGI->root = XCreateSimpleWindow( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                                   def.x, def.y, def.width, def.height, 0,
                                   WhitePixel(XBGI->dpy,XBGI->scr),
                                   BlackPixel(XBGI->dpy,XBGI->scr) );
  def.min_width  = 1;
  def.min_height = 1;
  def.max_width  = 2000;
  def.max_height = 2000;
  def.flags = PMinSize | PMaxSize;
  if ( flag_mask&XValue )      def.flags = def.flags | PPosition;
  if ( flag_mask&YValue )      def.flags = def.flags | USPosition;
  if ( flag_mask&WidthValue )  def.flags = def.flags | PSize;
  if ( flag_mask&HeightValue ) def.flags = def.flags | USSize;
  XSelectInput( XBGI->dpy, XBGI->root, ExposureMask|StructureNotifyMask );
  XSetStandardProperties( XBGI->dpy, XBGI->root,
                         titlename, iconname, None, argv, argc, &def );
  XSetWMNormalHints( XBGI->dpy, XBGI->root, &def );

  /* input event selection, GC creation and so on */
  XBGI->depth   = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->gcdraw  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcfill  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcpixel = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcimage = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  att.backing_store = Always;
  if ( scan_option( argc, argv, "-backing"  ) )
    att.backing_store = Always;
  if ( scan_option( argc, argv, "-backing-" ) )
    att.backing_store = NotUseful;
  XBGI->win = (Window*)malloc( XBGI->npage*sizeof(Window) );
  for ( i=0; i<XBGI->npage; i++ ){
    XBGI->win[i] = XCreateSimpleWindow( XBGI->dpy, XBGI->root, 0, 0,
                                       def.width, def.height, 0,
                                       WhitePixel(XBGI->dpy,XBGI->scr),
                                       BlackPixel(XBGI->dpy,XBGI->scr) );
    XSelectInput( XBGI->dpy, XBGI->win[i], ExposureMask|
                 OwnerGrabButtonMask|ButtonPressMask|ButtonReleaseMask|
                 PointerMotionMask|ButtonMotionMask|KeyPressMask|
                 EnterWindowMask|LeaveWindowMask );
    XChangeWindowAttributes( XBGI->dpy, XBGI->win[i], CWBackingStore, &att );
    XSetStandardProperties( XBGI->dpy, XBGI->win[i],
                           titlename, iconname, None, argv, argc, &def );
  }
  
  /* initializing fonts */
  for ( i=DEFAULT_FONT; i<=MAX_CHAR_FONT; i++ )
    for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ )
      XBGI->chr[i][j] = NULL;
#ifdef ENABLE_BILINGUAL
  for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ )
    XBGI->bchr0[j] = NULL;
#endif
#ifdef ENABLE_MULTILINGUAL
  for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ )
    XBGI->mchr0[j] = NULL;
#endif

  /* defualt event procedures in xbgiloop */
  XBGI->button_press_handler   = BGI_button_press;
  XBGI->button_release_handler = BGI_button_release;
  XBGI->mouse_motion_handler   = BGI_mouse_motion;
  XBGI->page_enter_handler     = BGI_page_enter;
  XBGI->page_leave_handler     = BGI_page_leave;
  XBGI->key_press_handler      = BGI_key_press;
  XBGI->configure_handler      = BGI_configure;
  XBGI->expose_handler         = BGI_expose;
  XBGI->noevent_handler        = BGI_noevent;

  /* BGI default setting */
  detectgraph( &(XBGI->gdriver), &(XBGI->gmode) );
  setgraphmode( XBGI->gmode );
  graphdefaults();

  XMapWindow( XBGI->dpy, XBGI->root );
  XMapSubwindows( XBGI->dpy, XBGI->root ); 
  for ( i=0; i<XBGI->npage; i++ )
    do {
      XNextEvent( XBGI->dpy, &ev );
    } while ( Expose != ev.type );
  XFlush( XBGI->dpy );

  /* set icon from data or file */
  if ( strcmp(iconfile,"")==0 ){
    for ( i=0; xbms[i].bits != NULL; i++ ){
      bm = XCreateBitmapFromData( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                                 xbms[i].bits, xbms[i].width, xbms[i].height );
      status = XSetIconFromPixmap( XBGI->dpy, XBGI->root, bm,
                                  (unsigned int)xbms[i].width,
                                  (unsigned int)xbms[i].height );
      XFreePixmap( XBGI->dpy, bm );
      if ( status ) break;
    }
  }else{
    seticon( iconfile );
  }

  XBGI->gerror = grOk;
  XBGI->_driver = _BGI_driver;
}

int BGI_driver_far[] = { 999 }; /* dummy */
