/*
 * Source File < graphix.c >
 *
 * Enhanced Functions: Borland Graphics Interface for Xlib
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#include "_bgi.h"

#define XBGIENVIRON "XBGISET"
#define XBGISETFILE ".xbgisets"
#define SPACES " \t\n"
#define iscomment(c) ((c)=='#')

xbgi *_creategraph( int argc, char *argv[], char *pd )
{
  static int id = 0;
  xbgi *this;
  int i, j;
  char buf[256], *ptr;
  
  this = (xbgi*)malloc(sizeof(xbgi));
  if ( !this ) FatalError( "memory allocation in _creategraph()" );
  /*
   * insert to linked list
   */
  this->next = XBGIS;
  XBGIS = this;
  /*
   * set to current XBGI
   */
  XBGI = this;
  /*
   * set ID
   */
  XBGI->id = id++;
  /*
   * set to parameter options
   */
  XBGI->argc = argc;
  XBGI->argv = (char**)malloc( argc*sizeof(char*) );
  if ( !XBGI->argv ) FatalError( "memory allocation in _creategraph()" );
  for ( i=0; i<argc; i++ ){
    XBGI->argv[i] = (char*)malloc( (strlen(argv[i])+1)*sizeof(char) );
    strcpy( XBGI->argv[i], argv[i] );
  }
  /*
   * set to parameter options of pd
   */
  strcpy( buf, pd );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    XBGI->argc++;
    ptr = strtok( NULL, SPACES );
  }
  if ( XBGI->argv )
    XBGI->argv = (char**)realloc( XBGI->argv, XBGI->argc*sizeof(char*) );
  else
    XBGI->argv = (char**)malloc( XBGI->argc*sizeof(char*) );
  if ( !XBGI->argv ) FatalError( "memory allocation in _creategraph()" );
  strcpy( buf, pd );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    XBGI->argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
    strcpy( XBGI->argv[argc++], ptr );
    ptr = strtok( NULL, SPACES );
  }
  /*
   * xbgi system initializing by driver
   */
  XBGI->gerror  = grNoInitGraph;
  XBGI->conio   = NULL;
  XBGI->driver  = XBGI_CURRENT_DRIVER;
  XBGI->_driver = NULL;
  /*
   * call driver
   */
  XBGI->driver();
  return XBGI;
}

xbgi *creategraph( int *graphdriver, int *graphmode, char *pathtodriver )
{
  int argc;
  char **argv;
  int i, j;
  char *ptr, options[256], buf[1024], buf2[256], file[64], key[64];
  FILE *fp;
  xbgi *this;

  /*
   * parameter settings to argc, argv from xbgienviron
   */
  sprintf( options, "%s -gd %d -gm %d ", "xbgi", *graphdriver, *graphmode );
  if ( getenv( XBGIENVIRON ) != NULL ){
    strcat( options, getenv( XBGIENVIRON ) );
    strcat( options, " " );
  }
  strcat( options, pathtodriver );
  strcpy( buf, options );
  ptr = strtok( buf, SPACES );
  argc = 0;
  while ( ptr != NULL ){
    argc++;
    ptr = strtok( NULL, SPACES );
  }
  argv = (char**)malloc( argc*sizeof(char*) );
  if ( !argv ) FatalError( "memory allocation in creategraph()" );
  i = 0; 
  strcpy( buf, options );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
    strcpy( argv[i++], ptr );
    ptr = strtok( NULL, SPACES );
  }
  /*
   * parameter settings to argc, argv from xbgisetfile
   */
  strcpy( file, "" );
  ptr = getenv( "HOME" );
  if ( ptr != NULL ){
    strcpy( file, ptr );
    strcat( file, "/" );
    strcat( file, XBGISETFILE );
  }
  if ( (fp=fopen(file,"r")) != NULL ){
    j = i = strspn( pathtodriver, SPACES );
    if ( strcmp(pathtodriver,"") == 0 || pathtodriver[i] == '-' ){
      strcpy( key, "" );
    }else{
      while( pathtodriver[j]!=' ' && pathtodriver[j]!='\0' ) j++;
      strncat( key, &pathtodriver[i], j-i+1 );
    }
    while ( fgets( buf, 256, fp ) != NULL ){
      i = strspn( buf, SPACES );
      if ( i==strlen( buf ) || iscomment( buf[i] ) )
        continue;
      while ( buf[strlen(buf)-2] == '\\' ){
        buf[strlen(buf)-2] = '\0';
        if ( fgets( buf2, 256, fp ) != NULL ){
          if ( strlen(buf)+strlen(buf2) < 1024 )
            strcat( buf, buf2 );
        }else{
          break;
        }
      }
      i = strspn( buf, SPACES );
      if ( ( strcmp( key, "" )==0 && buf[i] == '-' )
          || ( strncmp(&buf[i],key,strlen(key)) == 0 ) ){
        i = argc;
        strcpy( options, buf );
        ptr = strtok( options, SPACES );
        while ( ptr != NULL ){
          argc++;
          ptr = strtok( NULL, SPACES );
        }
        argv = (char**)realloc( argv, argc*sizeof(char*) );
        if ( !argv ) FatalError( "memory allocation in creategraph()" );
        strcpy( options, buf );
        ptr = strtok( options, SPACES );
        while ( ptr != NULL ){
          argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
          strcpy( argv[i++], ptr );
          ptr = strtok( NULL, SPACES );
        }
        break;
      }
    }
    fclose( fp );
  }
  /*
   * create xbgi
   */
  this = _creategraph( argc, argv, "" );
  for ( i=0; i<argc; i++ )
    free( argv[i] );
  free ( argv );
  *graphdriver = this->gdriver;
  *graphmode   = this->gmode;
  return this;
}

xbgi *thisgraph( void )
{
  return XBGI;
}

void currentgraph( xbgi *this )
{
  XBGI = this;
}

void destroygraph( xbgi *this )
{
  int i;
  xbgi *ptr;

  /*
   * call destroyer of driver
   */
  if ( this->_driver )
    this->_driver();
  /*
   * free memory
   */
  for ( i=0; i<this->argc; i++ )
    free( this->argv[i] );
  free ( this->argv );
  /*
   * delete from linked list
   */
  if ( XBGIS == this ){
    XBGIS = this->next;
    free( this );
    this = NULL;
    return;
  }
  ptr = XBGIS;
  while ( ptr->next != NULL ){
    if ( ptr->next == this ){
      ptr->next = this->next;
      XBGI = ptr;
      free( this );
      this = NULL;
      return;
    }
    ptr = ptr->next;
  }
  return;
}
